/*
 * (c) 2014 Eon S. Jeon <esjeon@hyunmu.am>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

var Shiki = {}
if(module && module.exports)
  module.exports = Shiki;

/* A template for template function */
Shiki.template = (function template(locals) {
  var $buf$ = "", $line$ = 0;
  "%library%"

  try {
    with(locals || {})
      (function() { "%body%"
      }).call(this);
  } catch(e) {
    e.message = '%file%:' + $line$ + '\n\n' + e.message;
    throw e;
  }
  return $buf$;
}).toString().replace(/^[^{]*{\n/, '').replace(/\n\s*}$/, '')


/* Functions to be embedded into template function */
Shiki.library = {
  'print': function print(obj) {
    $buf$ += obj.toString();
  },
}


/* Parse the given template code into template function body */
Shiki.parse = function(str, opts) {
  var parts = []
    , line = 0
  var rest = str
    , match
    , code, expr, text

  while((match = rest.match(/<%(([^%]|%[^>])*)%>|\n?^\s*%(.*)$/m))) {
    expr = match[1]
    code = match[3]
    text = rest.slice(0, match.index)

    parts.push({ type: 'text', str: text, line: line })
    line += (text.match(/\n/g) || []).length

    if(expr)
      parts.push({ type: 'expr', str: expr.trim(), line: line })
    else /* if(code) */ {
      code = code
      parts.push({ type: 'code', str: code.trim(), line: line })
    }

    rest = rest.slice(match.index + match[0].length)
  }

  if(rest.length > 0)
    parts.push({ type: 'text', str: rest, line: line })

  return parts.map(function(part) {
    switch(part.type) {
    case 'code': return '$line$=' + part.line + '; ' + part.str
    case 'text': return '$buf$+=' + JSON.stringify(part.str) + ';'
    case 'expr': return '$line$=' + part.line + ';' + '$buf$+=' + part.str+ ';'
    default: return ''
    }
  }).join('\n          ')
}

Shiki.compile = function(str, opts) {
  opts = opts || {}
  var filename = opts.filename || 'shiki'

  var body = Shiki.parse(str, opts)
  
  var lib = Object.keys(Shiki.library).map(function(name) {
    return Shiki.library[name].toString()
  }).join('\n')

  var fn = new Function('locals',  
        Shiki.template
          .replace('"%body%"', body)
          .replace('"%library%"', lib)
          .replace('%file%', filename)
      )

  return fn
}

